use proc_macro::TokenStream;
use quote::{format_ident, quote};
use syn::{FnArg, Ident, ItemFn, Pat, ReturnType, Type, parse_macro_input};

/// This macro turns this function into a worker.
///
/// This will upgrade the function and generate a few ones (`fn` is a placeholder for the functions name):
/// - `fn()` - Behaves exactly the same as the original but it will be computed by a worker.
/// - `fn_init(ServiceManager) -> ServiceManager` - Registers worker threads on a `ServiceManager`.
/// - `fn_shutdown()` - Issue a shutdown request.
/// - `fn_init_scoped(ServiceManager) -> (ServiceManager, fn_Scoped)` - Registers worker threads on a `ServiceManager` and return a scoped struct. You can call the underlying function with `.call()` on the struct and it will automatically shutdown any workers if it gets out of scope.
/// - `fn_init_union(ServiceManager) -> ServiceManager` - Register worker threads and work in a union using Valkey.
/// - `fn_init_union_scoped(ServiceManager) -> ServiceManager` - Register worker threads and work in a union using Valkey and return a scoped struct.
/// - `fn_register_union()` - Register the worker in a union setup without starting local worker threads.
///
/// # Examples
/// ```ignore
/// use comrade::worker;
///
/// // Declare worker
/// #[worker]
/// pub fn multiply(a: i32, b: i32) -> i32 {
///    a * b
/// }
///
/// fn main() {
///    let s = ServiceManager::new().mode(comrade::service::ServiceMode::Decay);
///    
///    // Init worker thread
///    let s = multiply_init(s);
///    let s = s.spawn();
///
///    // Usage
///    let x = multiply(2, 8);
///    println!("myfn {x}");
///
///    // Shutdown worker thread
///    multiply_shutdown();
///
///    s.join().unwrap();
/// }
/// ```
#[proc_macro_attribute]
pub fn worker(attr: TokenStream, item: TokenStream) -> TokenStream {
    let input: ItemFn = parse_macro_input!(item as ItemFn);
    let fn_name = &input.sig.ident;

    // Parse optional attribute argument (e.g., #[worker(4)])
    let worker_count = attr.to_string().parse::<usize>().unwrap_or(1);

    // Extract parameter names and types separately
    let params: Vec<(Ident, Type)> = input
        .sig
        .inputs
        .iter()
        .filter_map(|arg| {
            if let FnArg::Typed(pat_type) = arg {
                let name = if let Pat::Ident(pat_ident) = *pat_type.pat.clone() {
                    pat_ident.ident.clone()
                } else {
                    return None;
                };
                let ty = *pat_type.ty.clone();
                Some((name, ty))
            } else {
                None
            }
        })
        .collect();

    // Extract parameter names and types into separate lists for quoting
    let param_names: Vec<Ident> = params.iter().map(|(name, _)| name.clone()).collect();
    let param_types: Vec<Type> = params.iter().map(|(_, ty)| ty.clone()).collect();

    // Extract return type
    let return_type = match &input.sig.output {
        ReturnType::Type(_, ty) => quote!(#ty),
        ReturnType::Default => quote!(()),
    };

    // Extract function body
    let body = &input.block;

    let wrapper_fn = format_ident!("{}_wrapper", fn_name);
    let worker_fn = format_ident!("{}_worker", fn_name);
    let init_fn = format_ident!("{}_init", fn_name);
    let init_fn_union = format_ident!("{}_init_union", fn_name);
    let init_fn_scoped = format_ident!("{}_init_scoped", fn_name);
    let init_fn_scoped_union = format_ident!("{}_init_union_scoped", fn_name);
    let fn_scope_struct = format_ident!("{}_Scoped", fn_name);
    let fn_name_async = format_ident!("{}_async", fn_name);
    let init_fn_register = format_ident!("{}_register_union", fn_name);
    let shutdown_fn = format_ident!("{}_shutdown", fn_name);

    let param_unpacking = param_names.iter().enumerate().map(|(i, name)| {
        if param_names.len() == 1 {
            return quote! {
                let #name = i;

            };
        }

        let param_type = &param_types[i];
        if let Type::Path(_) = param_type {
            quote! {
                let #name = i.#i;
            }
        } else {
            quote! {
               let #name = i;
            }
        }
    });

    let output = quote! {
        pub fn #fn_name(#(#param_names: #param_types),*) -> #return_type {
            let i: comrade::serde_json::Value = comrade::serde_json::to_value( (#(#param_names),*) ).unwrap();
            serde_json::from_value(comrade::UNION.get(stringify!(#fn_name)).expect("Function call is not registered in UNION").send(i)).unwrap()
        }

        #[doc = "Will run the function non blocking returning a `JobResult<_>` for fetching a result later."]
        pub fn #fn_name_async(#(#param_names: #param_types),*) -> comrade::job::JobResult<comrade::serde_json::Value> {
            let i: comrade::serde_json::Value = comrade::serde_json::to_value( (#(#param_names),*) ).unwrap();
            if let Some(dispatch) = comrade::UNION.get(stringify!(#fn_name)) {
                dispatch.send_async(i)
            } else {
                panic!("Function call {} is not registered in UNION", stringify!(#fn_name))
            }
        }

        fn #wrapper_fn(task: JobOrder<comrade::serde_json::Value, comrade::serde_json::Value>) {
            let i = task.param.clone();

            // Deserialize the parameter into the function's expected types
            let i: (#(#param_types),*) = comrade::serde_json::from_value(i).unwrap();

            #(#param_unpacking)*

            let res = #body;

            task.done(comrade::serde_json::to_value(&res).unwrap());
        }

        pub fn #worker_fn(recv: comrade::job::TaskReceiverBackend<comrade::serde_json::Value, comrade::serde_json::Value>) {
            let mut metrics = (0, 0);
            loop {
                let task = recv.recv();

                match task {
                    Ok(task) => {
                        if let comrade::serde_json::Value::Object(obj) = &task.param {
                            if obj.contains_key("task") {
                                log::info!("Shutdown requested for task worker {}. Processed {} tasks since start with {} errors.", stringify!(#fn_name), metrics.0, metrics.1);
                                task.done(comrade::serde_json::json!({"ok": 1}));
                                break;
                            }
                        }

                        #wrapper_fn(task);
                        metrics.0 += 1;
                    },
                    Err(e) => {
                        metrics.1 += 1;
                        log::error!("Error receiving task: {e:?}");
                    }
                }
            }
        }

        #[doc = "Shutdown the worker"]
        pub fn #shutdown_fn() {
            comrade::UNION.get(stringify!(#fn_name)).unwrap().send_all(comrade::serde_json::json!({"task": "shutdown"}));
        }

        #[doc = "Initialize a worker thread on `ServiceManager`"]
        pub fn #init_fn(sm: ServiceManager) -> ServiceManager {
            let mut dispatchers = Vec::new();
            let mut s = sm;

            log::info!("Initializing worker {} with {} threads", stringify!(#worker_fn), #worker_count);

            for i in 0..#worker_count {
                let (dispatch, recv): (JobDispatcher<_, _>, Receiver<JobOrder<_, _>>) = JobDispatcher::new();

                s = s.register(
                        &format!("{}_{i}", stringify!(#worker_fn)),
                        move |_| {
                            #worker_fn(comrade::job::TaskReceiverBackend::Local(recv.clone()))
                        }
                    );

                    dispatchers.push(comrade::job::Dispatcher::Local(dispatch));
                }

            comrade::UNION.insert(stringify!(#fn_name), comrade::job::JobMultiplexer::from(dispatchers));

            s
        }


        #[doc = "Register worker union on Valkey backend without starting local workers."]
        pub fn #init_fn_register() {
            let mut dispatchers = Vec::new();

            log::info!("Registering worker union {}", stringify!(#worker_fn));

            let dispatch = comrade::job::ValkeyJobDispatcher::<_, _>::new_topic(stringify!(#worker_fn), false);
            dispatchers.push(comrade::job::Dispatcher::Union(dispatch));

            comrade::UNION.insert(stringify!(#fn_name), comrade::job::JobMultiplexer::from(dispatchers));
        }

        #[doc = "Initialize worker threads on `ServiceManager` with Valkey backend"]
        pub fn #init_fn_union(sm: ServiceManager) -> ServiceManager {
            let mut dispatchers = Vec::new();
            let mut s = sm;

            log::info!("Initializing worker union {} with {} threads", stringify!(#worker_fn), #worker_count);

            for i in 0..#worker_count {
                let dispatch = comrade::job::ValkeyJobDispatcher::<_, _>::new_topic(stringify!(#worker_fn), true);

                let recv = dispatch.clone();

                s = s.register(
                        &format!("{}_union_{i}", stringify!(#worker_fn)),
                        move |_| {
                            #worker_fn(comrade::job::TaskReceiverBackend::Union(recv.clone()))
                        }
                    );

                    dispatchers.push(comrade::job::Dispatcher::Union(dispatch));
                }

            comrade::UNION.insert(stringify!(#fn_name), comrade::job::JobMultiplexer::from(dispatchers));

            s
        }

        #[allow(non_camel_case_types)]
        pub struct #fn_scope_struct {}

        impl #fn_scope_struct {
            pub fn call(&self, #(#param_names: #param_types),*) -> #return_type {
                #fn_name(#(#param_names),*)
            }
        }

        impl Drop for #fn_scope_struct {
            fn drop(&mut self) {
                log::info!("Scoped task worker got dropped.");
                #shutdown_fn();
            }
        }

        #[doc = "Initialize a worker thread on `ServiceManager` on a scoped lifetime"]
        pub fn #init_fn_scoped(sm: ServiceManager) -> (ServiceManager, #fn_scope_struct) {
            let sm = #init_fn(sm);
            (sm, #fn_scope_struct {})
        }

        #[doc = "Initialize a worker union on `ServiceManager` on a scoped lifetime"]
        pub fn #init_fn_scoped_union(sm: ServiceManager) -> (ServiceManager, #fn_scope_struct) {
            let sm = #init_fn_union(sm);
            (sm, #fn_scope_struct {})
        }
    };

    output.into()
}
